/*
 * Copyright 2019, OpenTelemetry Authors
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package io.opentelemetry.sdk.trace.export;

import static com.google.common.truth.Truth.assertThat;
import static org.mockito.Mockito.doThrow;

import io.opentelemetry.sdk.common.export.ConfigBuilder;
import io.opentelemetry.sdk.trace.ReadableSpan;
import io.opentelemetry.sdk.trace.Samplers;
import io.opentelemetry.sdk.trace.TestUtils;
import io.opentelemetry.sdk.trace.TracerSdkProvider;
import io.opentelemetry.sdk.trace.data.SpanData;
import io.opentelemetry.trace.Tracer;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicBoolean;
import javax.annotation.Nullable;
import javax.annotation.concurrent.GuardedBy;
import org.junit.After;
import org.junit.Before;
import org.junit.Rule;
import org.junit.Test;
import org.junit.contrib.java.lang.system.EnvironmentVariables;
import org.junit.contrib.java.lang.system.ProvideSystemProperty;
import org.junit.runner.RunWith;
import org.junit.runners.JUnit4;
import org.mockito.ArgumentMatchers;
import org.mockito.Mock;
import org.mockito.MockitoAnnotations;

/** Unit tests for {@link BatchSpansProcessor}. */
@RunWith(JUnit4.class)
public class BatchSpansProcessorTest {

  private static final String SPAN_NAME_1 = "MySpanName/1";
  private static final String SPAN_NAME_2 = "MySpanName/2";
  private static final long MAX_SCHEDULE_DELAY_MILLIS = 500;
  private final TracerSdkProvider tracerSdkFactory = TracerSdkProvider.builder().build();
  private final Tracer tracer = tracerSdkFactory.get("BatchSpansProcessorTest");
  private final BlockingSpanExporter blockingSpanExporter = new BlockingSpanExporter();
  @Mock private SpanExporter mockServiceHandler;

  abstract static class ConfigTest extends BatchSpansProcessor.Config.Builder {

    public static ConfigBuilder.NamingConvention getDot() {
      return NamingConvention.DOT;
    }

    public static ConfigBuilder.NamingConvention getEnv() {
      return NamingConvention.ENV_VAR;
    }
  }

  @Before
  public void setup() {
    MockitoAnnotations.initMocks(this);
  }

  @After
  public void cleanup() {
    tracerSdkFactory.shutdown();
  }

  @RunWith(JUnit4.class)
  public static class ConfigurationSystemPropertiesTest {
    @Rule
    public final ProvideSystemProperty systemProperties =
        new ProvideSystemProperty("otel.bsp.schedule.delay", "5")
            .and("otel.bsp.max.queue", "5")
            .and("otel.bsp.export.timeout", "5")
            .and("otel.bsp.export.sampled", "false");

    @Test
    public void testSystemProperties() {
      BatchSpansProcessor.Config config = BatchSpansProcessor.Config.loadFromDefaultSources();
      assertThat(config.getScheduleDelayMillis()).isEqualTo(5);
      assertThat(config.getMaxQueueSize()).isEqualTo(5);
      // This is not defined and should be equal to the default values
      assertThat(config.getMaxExportBatchSize())
          .isEqualTo(BatchSpansProcessor.Config.getDefault().getMaxExportBatchSize());
      assertThat(config.getExporterTimeoutMillis()).isEqualTo(5);
      assertThat(config.isExportOnlySampled()).isEqualTo(false);
    }
  }

  @RunWith(JUnit4.class)
  public static class ConfigurationEnvironmentVariablesTest {
    @Rule
    public final EnvironmentVariables environmentVariables =
        new EnvironmentVariables()
            .set("OTEL_BSP_MAX_QUEUE", "22")
            .set("OTEL_BSP_EXPORT_TIMEOUT", "22")
            .set("OTEL_BSP_EXPORT_SAMPLED", "true");

    @Test
    public void testEnvironmentVariables() {
      BatchSpansProcessor.Config config = BatchSpansProcessor.Config.loadFromDefaultSources();
      BatchSpansProcessor.Config defaultConf = BatchSpansProcessor.Config.getDefault();
      // This is not defined and should be equal to the default values
      assertThat(config.getScheduleDelayMillis()).isEqualTo(defaultConf.getScheduleDelayMillis());
      assertThat(config.getMaxQueueSize()).isEqualTo(22);
      // This is not defined and should be equal to the default values
      assertThat(config.getMaxExportBatchSize()).isEqualTo(defaultConf.getMaxExportBatchSize());
      assertThat(config.getExporterTimeoutMillis()).isEqualTo(22);
      assertThat(config.isExportOnlySampled()).isEqualTo(true);
    }
  }

  @RunWith(JUnit4.class)
  public static class ConfigurationSystemAndEnvironmentTest {
    @Rule
    public final ProvideSystemProperty systemProperties =
        new ProvideSystemProperty("otel.bsp.max.queue", "5")
            .and("otel.bsp.export.timeout", "5")
            .and("otel.bsp.export.sampled", "false");

    @Rule
    public final EnvironmentVariables environmentVariables =
        new EnvironmentVariables()
            .set("OTEL_BSP_SCHEDULE_DELAY", "22")
            .set("OTEL_BSP_MAX_QUEUE", "22")
            .set("OTEL_BSP_EXPORT_TIMEOUT", "22")
            .set("OTEL_BSP_EXPORT_SAMPLED", "true");

    @Test
    public void testSystemAndEnv() {
      BatchSpansProcessor.Config config = BatchSpansProcessor.Config.loadFromDefaultSources();
      BatchSpansProcessor.Config defaultConf = BatchSpansProcessor.Config.getDefault();
      assertThat(config.getScheduleDelayMillis()).isEqualTo(22);
      assertThat(config.getMaxQueueSize()).isEqualTo(5);
      // This is not defined and should be equal to the default values
      assertThat(config.getMaxExportBatchSize()).isEqualTo(defaultConf.getMaxExportBatchSize());
      assertThat(config.getExporterTimeoutMillis()).isEqualTo(5);
      assertThat(config.isExportOnlySampled()).isEqualTo(false);
    }
  }

  private ReadableSpan createSampledEndedSpan(String spanName) {
    io.opentelemetry.trace.Span span =
        TestUtils.startSpanWithSampler(tracerSdkFactory, tracer, spanName, Samplers.alwaysOn())
            .startSpan();
    span.end();
    return (ReadableSpan) span;
  }

  @Test
  public void testConfiguration() {
    BatchSpansProcessor.Config config;
    BatchSpansProcessor.Config defConfig = BatchSpansProcessor.Config.getDefault();

    config = BatchSpansProcessor.Config.newBuilder().build();
    // check defaults
    assertThat(config.getScheduleDelayMillis()).isEqualTo(defConfig.getScheduleDelayMillis());
    assertThat(config.getMaxQueueSize()).isEqualTo(defConfig.getMaxQueueSize());
    assertThat(config.getMaxExportBatchSize()).isEqualTo(defConfig.getMaxExportBatchSize());
    assertThat(config.getExporterTimeoutMillis()).isEqualTo(defConfig.getExporterTimeoutMillis());
    assertThat(config.isExportOnlySampled()).isEqualTo(defConfig.isExportOnlySampled());

    // check system configuration
    Map<String, String> configMap = new HashMap<>();
    configMap.put("otel.bsp.schedule.delay", "1");
    configMap.put("otel.bsp.max.queue", "1");
    configMap.put("otel.bsp.max.export.batch", "1");
    configMap.put("otel.bsp.export.timeout", "1");
    configMap.put("otel.bsp.export.sampled", "false");
    config =
        BatchSpansProcessor.Config.newBuilder()
            .fromConfigMap(configMap, ConfigTest.getDot())
            .build();
    assertThat(config.getScheduleDelayMillis()).isEqualTo(1);
    assertThat(config.getMaxQueueSize()).isEqualTo(1);
    assertThat(config.getMaxExportBatchSize()).isEqualTo(1);
    assertThat(config.getExporterTimeoutMillis()).isEqualTo(1);
    assertThat(config.isExportOnlySampled()).isEqualTo(false);

    // check properties
    Properties properties = new Properties();
    for (Map.Entry<String, String> entry : configMap.entrySet()) {
      properties.put(entry.getKey(), entry.getValue());
    }
    config = BatchSpansProcessor.Config.newBuilder().readProperties(properties).build();
    assertThat(config.getScheduleDelayMillis()).isEqualTo(1);
    assertThat(config.getMaxQueueSize()).isEqualTo(1);
    assertThat(config.getMaxExportBatchSize()).isEqualTo(1);
    assertThat(config.getExporterTimeoutMillis()).isEqualTo(1);
    assertThat(config.isExportOnlySampled()).isEqualTo(false);

    // Check env vars
    configMap.clear();
    configMap.put("OTEL_BSP_SCHEDULE_DELAY", "2");
    configMap.put("OTEL_BSP_MAX_QUEUE", "2");
    configMap.put("OTEL_BSP_MAX_EXPORT_BATCH", "2");
    configMap.put("OTEL_BSP_EXPORT_TIMEOUT", "2");
    configMap.put("OTEL_BSP_EXPORT_SAMPLED", "true");
    config =
        BatchSpansProcessor.Config.newBuilder()
            .fromConfigMap(configMap, ConfigTest.getEnv())
            .build();
    assertThat(config.getScheduleDelayMillis()).isEqualTo(2);
    assertThat(config.getMaxQueueSize()).isEqualTo(2);
    assertThat(config.getMaxExportBatchSize()).isEqualTo(2);
    assertThat(config.getExporterTimeoutMillis()).isEqualTo(2);
    assertThat(config.isExportOnlySampled()).isEqualTo(true);

    // Check mixing conventions
    configMap.clear();
    configMap.put("OTEL_BSP_schedule_DELAY", "3");
    configMap.put("OTEL.BSP.MAX_QUEUE", "3");
    configMap.put("OTEL_bsp.MAX_EXPORT_BATCH", "3");
    configMap.put("OTEL_BSP_ExPoRt_TIMEOUT", "3");
    configMap.put("OTEL_bSp.EXPORT.SAmpleD", "false");
    config =
        BatchSpansProcessor.Config.newBuilder()
            .fromConfigMap(configMap, ConfigTest.getEnv())
            .build();
    assertThat(config.getScheduleDelayMillis()).isEqualTo(3);
    assertThat(config.getMaxQueueSize()).isEqualTo(3);
    assertThat(config.getMaxExportBatchSize()).isEqualTo(3);
    assertThat(config.getExporterTimeoutMillis()).isEqualTo(3);
    assertThat(config.isExportOnlySampled()).isEqualTo(false);
  }

  @Test
  public void testOnlySetPropertiesOverrideDefaults() {
    BatchSpansProcessor.Config config;
    Map<String, String> configMap = new HashMap<>();
    // check only set values are written
    configMap.clear();
    configMap.put("OTEL_BSP_SCHEDULE_DELAY", "1");
    config =
        BatchSpansProcessor.Config.newBuilder()
            .fromConfigMap(configMap, ConfigTest.getEnv())
            .build();
    assertThat(config.getScheduleDelayMillis()).isEqualTo(1);
    assertThat(config.getMaxQueueSize()).isEqualTo(2048);
    assertThat(config.getMaxExportBatchSize()).isEqualTo(512);
    assertThat(config.getExporterTimeoutMillis()).isEqualTo(30_000);
    assertThat(config.isExportOnlySampled()).isEqualTo(true);
  }

  @Test
  public void testUserSetPropertiesOverrideDefaults() {
    BatchSpansProcessor.Config config;
    Map<String, String> configMap = new HashMap<>();
    // check only set values are written
    configMap.clear();
    configMap.put("otel.bsp.schedule.delay", "1");
    configMap.put("otel.bsp.max.queue", "1");
    configMap.put("otel.bsp.max.export.batch", "1");
    configMap.put("otel.bsp.export.timeout", "1");
    configMap.put("otel.bsp.export.sampled", "false");
    config =
        BatchSpansProcessor.Config.newBuilder()
            .fromConfigMap(configMap, ConfigTest.getEnv())
            .setScheduleDelayMillis(2)
            .setMaxQueueSize(2)
            .setMaxExportBatchSize(2)
            .setExporterTimeoutMillis(2)
            .setExportOnlySampled(true)
            .build();
    assertThat(config.getScheduleDelayMillis()).isEqualTo(2);
    assertThat(config.getMaxQueueSize()).isEqualTo(2);
    assertThat(config.getMaxExportBatchSize()).isEqualTo(2);
    assertThat(config.getExporterTimeoutMillis()).isEqualTo(2);
    assertThat(config.isExportOnlySampled()).isEqualTo(true);
  }

  // TODO(bdrutu): Fix this when Sampler return RECORD option.
  /*
  private ReadableSpan createNotSampledRecordingEventsEndedSpan(String spanName) {
    io.opentelemetry.trace.Span span =
        tracer.spanBuilder(spanName).setSampler(Samplers.neverSample()).startSpanWithSampler();
    span.end();
    return (ReadableSpan) span;
  }
  */

  private void createNotSampledEndedSpan(String spanName) {
    TestUtils.startSpanWithSampler(tracerSdkFactory, tracer, spanName, Samplers.alwaysOff())
        .startSpan()
        .end();
  }

  @Test
  public void startEndRequirements() {
    BatchSpansProcessor spansProcessor = BatchSpansProcessor.create(new WaitingSpanExporter(0));
    assertThat(spansProcessor.isStartRequired()).isFalse();
    assertThat(spansProcessor.isEndRequired()).isTrue();
  }

  @Test
  public void exportDifferentSampledSpans() {
    WaitingSpanExporter waitingSpanExporter = new WaitingSpanExporter(2);
    tracerSdkFactory.addSpanProcessor(
        BatchSpansProcessor.create(
            waitingSpanExporter,
            BatchSpansProcessor.Config.newBuilder()
                .setScheduleDelayMillis(MAX_SCHEDULE_DELAY_MILLIS)
                .build()));

    ReadableSpan span1 = createSampledEndedSpan(SPAN_NAME_1);
    ReadableSpan span2 = createSampledEndedSpan(SPAN_NAME_2);
    List<SpanData> exported = waitingSpanExporter.waitForExport();
    assertThat(exported).containsExactly(span1.toSpanData(), span2.toSpanData());
  }

  @Test
  public void exportMoreSpansThanTheBufferSize() {
    WaitingSpanExporter waitingSpanExporter = new WaitingSpanExporter(6);
    BatchSpansProcessor.Config config =
        BatchSpansProcessor.Config.newBuilder()
            .setMaxQueueSize(6)
            .setMaxExportBatchSize(2)
            .setScheduleDelayMillis(MAX_SCHEDULE_DELAY_MILLIS)
            .build();

    tracerSdkFactory.addSpanProcessor(BatchSpansProcessor.create(waitingSpanExporter, config));

    ReadableSpan span1 = createSampledEndedSpan(SPAN_NAME_1);
    ReadableSpan span2 = createSampledEndedSpan(SPAN_NAME_1);
    ReadableSpan span3 = createSampledEndedSpan(SPAN_NAME_1);
    ReadableSpan span4 = createSampledEndedSpan(SPAN_NAME_1);
    ReadableSpan span5 = createSampledEndedSpan(SPAN_NAME_1);
    ReadableSpan span6 = createSampledEndedSpan(SPAN_NAME_1);
    List<SpanData> exported = waitingSpanExporter.waitForExport();
    assertThat(exported)
        .containsExactly(
            span1.toSpanData(),
            span2.toSpanData(),
            span3.toSpanData(),
            span4.toSpanData(),
            span5.toSpanData(),
            span6.toSpanData());
  }

  @Test
  public void forceExport() {
    WaitingSpanExporter waitingSpanExporter = new WaitingSpanExporter(1, 1);
    BatchSpansProcessor.Config config =
        BatchSpansProcessor.Config.newBuilder()
            .setMaxQueueSize(10_000)
            .setMaxExportBatchSize(2_000)
            .setScheduleDelayMillis(10_000) // 10s
            .build();
    BatchSpansProcessor batchSpansProcessor =
        BatchSpansProcessor.create(waitingSpanExporter, config);

    tracerSdkFactory.addSpanProcessor(batchSpansProcessor);
    for (int i = 0; i < 100; i++) {
      createSampledEndedSpan("notExported");
    }
    List<SpanData> exported = waitingSpanExporter.waitForExport();
    assertThat(exported).isNotNull();
    assertThat(exported.size()).isEqualTo(0);
    batchSpansProcessor.forceFlush();
    exported = waitingSpanExporter.waitForExport();
    assertThat(exported).isNotNull();
    assertThat(exported.size()).isEqualTo(100);
  }

  @Test
  public void exportSpansToMultipleServices() {
    WaitingSpanExporter waitingSpanExporter = new WaitingSpanExporter(2);
    WaitingSpanExporter waitingSpanExporter2 = new WaitingSpanExporter(2);
    BatchSpansProcessor.Config config =
        BatchSpansProcessor.Config.newBuilder()
            .setScheduleDelayMillis(MAX_SCHEDULE_DELAY_MILLIS)
            .build();
    tracerSdkFactory.addSpanProcessor(
        BatchSpansProcessor.create(
            MultiSpanExporter.create(
                Arrays.<SpanExporter>asList(waitingSpanExporter, waitingSpanExporter2)),
            config));

    ReadableSpan span1 = createSampledEndedSpan(SPAN_NAME_1);
    ReadableSpan span2 = createSampledEndedSpan(SPAN_NAME_2);
    List<SpanData> exported1 = waitingSpanExporter.waitForExport();
    List<SpanData> exported2 = waitingSpanExporter2.waitForExport();
    assertThat(exported1).containsExactly(span1.toSpanData(), span2.toSpanData());
    assertThat(exported2).containsExactly(span1.toSpanData(), span2.toSpanData());
  }

  @Test
  public void exportMoreSpansThanTheMaximumLimit() {
    final int maxQueuedSpans = 8;
    WaitingSpanExporter waitingSpanExporter = new WaitingSpanExporter(maxQueuedSpans);
    BatchSpansProcessor.Config config =
        BatchSpansProcessor.Config.newBuilder()
            .setScheduleDelayMillis(MAX_SCHEDULE_DELAY_MILLIS)
            .setMaxQueueSize(maxQueuedSpans)
            .setMaxExportBatchSize(maxQueuedSpans / 2)
            .build();
    tracerSdkFactory.addSpanProcessor(
        BatchSpansProcessor.create(
            MultiSpanExporter.create(Arrays.asList(blockingSpanExporter, waitingSpanExporter)),
            config));

    List<SpanData> spansToExport = new ArrayList<>(maxQueuedSpans + 1);
    // Wait to block the worker thread in the BatchSampledSpansProcessor. This ensures that no items
    // can be removed from the queue. Need to add a span to trigger the export otherwise the
    // pipeline is never called.
    spansToExport.add(createSampledEndedSpan("blocking_span").toSpanData());
    blockingSpanExporter.waitUntilIsBlocked();

    for (int i = 0; i < maxQueuedSpans; i++) {
      // First export maxQueuedSpans, the worker thread is blocked so all items should be queued.
      spansToExport.add(createSampledEndedSpan("span_1_" + i).toSpanData());
    }

    // TODO: assertThat(spanExporter.getReferencedSpans()).isEqualTo(maxQueuedSpans);

    // Now we should start dropping.
    for (int i = 0; i < 7; i++) {
      createSampledEndedSpan("span_2_" + i);
      // TODO: assertThat(getDroppedSpans()).isEqualTo(i + 1);
    }

    // TODO: assertThat(getReferencedSpans()).isEqualTo(maxQueuedSpans);

    // Release the blocking exporter
    blockingSpanExporter.unblock();

    // While we wait for maxQueuedSpans we ensure that the queue is also empty after this.
    List<SpanData> exported = waitingSpanExporter.waitForExport();
    assertThat(exported).isNotNull();
    assertThat(exported).containsExactlyElementsIn(spansToExport);
    exported.clear();
    spansToExport.clear();

    waitingSpanExporter.reset();
    // We cannot compare with maxReferencedSpans here because the worker thread may get
    // unscheduled immediately after exporting, but before updating the pushed spans, if that is
    // the case at most bufferSize spans will miss.
    // TODO: assertThat(getPushedSpans()).isAtLeast((long) maxQueuedSpans - maxBatchSize);

    for (int i = 0; i < maxQueuedSpans; i++) {
      spansToExport.add(createSampledEndedSpan("span_3_" + i).toSpanData());
      // No more dropped spans.
      // TODO: assertThat(getDroppedSpans()).isEqualTo(7);
    }

    exported = waitingSpanExporter.waitForExport();
    assertThat(exported).isNotNull();
    assertThat(exported).containsExactlyElementsIn(spansToExport);
  }

  @Test
  public void serviceHandlerThrowsException() {
    WaitingSpanExporter waitingSpanExporter = new WaitingSpanExporter(1);
    doThrow(new IllegalArgumentException("No export for you."))
        .when(mockServiceHandler)
        .export(ArgumentMatchers.<SpanData>anyList());

    BatchSpansProcessor.Config config =
        BatchSpansProcessor.Config.newBuilder()
            .setScheduleDelayMillis(MAX_SCHEDULE_DELAY_MILLIS)
            .build();
    tracerSdkFactory.addSpanProcessor(
        BatchSpansProcessor.create(
            MultiSpanExporter.create(Arrays.asList(mockServiceHandler, waitingSpanExporter)),
            config));
    ReadableSpan span1 = createSampledEndedSpan(SPAN_NAME_1);
    List<SpanData> exported = waitingSpanExporter.waitForExport();
    assertThat(exported).containsExactly(span1.toSpanData());
    waitingSpanExporter.reset();
    // Continue to export after the exception was received.
    ReadableSpan span2 = createSampledEndedSpan(SPAN_NAME_2);
    exported = waitingSpanExporter.waitForExport();
    assertThat(exported).containsExactly(span2.toSpanData());
  }

  @Test(timeout = 5000)
  public void exporterTimesOut() throws Exception {
    final CountDownLatch interruptMarker = new CountDownLatch(1);
    WaitingSpanExporter waitingSpanExporter =
        new WaitingSpanExporter(1) {
          @Override
          public ResultCode export(Collection<SpanData> spans) {
            ResultCode result = super.export(spans);
            try {
              // sleep longer than the configured timout of 100ms
              Thread.sleep(1000);
            } catch (InterruptedException e) {
              interruptMarker.countDown();
            }
            return result;
          }
        };

    int exporterTimeoutMillis = 100;
    BatchSpansProcessor.Config config =
        BatchSpansProcessor.Config.newBuilder()
            .setExporterTimeoutMillis(exporterTimeoutMillis)
            .setScheduleDelayMillis(1)
            .setMaxQueueSize(1)
            .build();

    tracerSdkFactory.addSpanProcessor(BatchSpansProcessor.create(waitingSpanExporter, config));

    ReadableSpan span = createSampledEndedSpan(SPAN_NAME_1);
    List<SpanData> exported = waitingSpanExporter.waitForExport();
    assertThat(exported).containsExactly(span.toSpanData());

    // since the interrupt happens outside the execution of the test method, we'll block to make
    // sure that the thread was actually interrupted due to the timeout.
    interruptMarker.await();
  }

  @Test
  public void exportNotSampledSpans() {
    WaitingSpanExporter waitingSpanExporter = new WaitingSpanExporter(1);
    BatchSpansProcessor.Config config =
        BatchSpansProcessor.Config.newBuilder()
            .setScheduleDelayMillis(MAX_SCHEDULE_DELAY_MILLIS)
            .build();
    tracerSdkFactory.addSpanProcessor(BatchSpansProcessor.create(waitingSpanExporter, config));

    createNotSampledEndedSpan(SPAN_NAME_1);
    createNotSampledEndedSpan(SPAN_NAME_2);
    ReadableSpan span2 = createSampledEndedSpan(SPAN_NAME_2);
    // Spans are recorded and exported in the same order as they are ended, we test that a non
    // sampled span is not exported by creating and ending a sampled span after a non sampled span
    // and checking that the first exported span is the sampled span (the non sampled did not get
    // exported).
    List<SpanData> exported = waitingSpanExporter.waitForExport();
    // Need to check this because otherwise the variable span1 is unused, other option is to not
    // have a span1 variable.
    assertThat(exported).containsExactly(span2.toSpanData());
  }

  @Test
  public void exportNotSampledSpans_recordingEvents() {
    // TODO(bdrutu): Fix this when Sampler return RECORD option.
    /*
    tracerSdkFactory.addSpanProcessor(
        BatchSpansProcessor.newBuilder(waitingSpanExporter)
            .setScheduleDelayMillis(MAX_SCHEDULE_DELAY_MILLIS)
            .reportOnlySampled(false)
            .build());

    ReadableSpan span = createNotSampledRecordingEventsEndedSpan(SPAN_NAME_1);
    List<SpanData> exported = waitingSpanExporter.waitForExport(1);
    assertThat(exported).containsExactly(span.toSpanData());
    */
  }

  @Test
  public void exportNotSampledSpans_reportOnlySampled() {
    // TODO(bdrutu): Fix this when Sampler return RECORD option.
    /*
    tracerSdkFactory.addSpanProcessor(
        BatchSpansProcessor.newBuilder(waitingSpanExporter)
            .reportOnlySampled(true)
            .setScheduleDelayMillis(MAX_SCHEDULE_DELAY_MILLIS)
            .build());

    createNotSampledRecordingEventsEndedSpan(SPAN_NAME_1);
    ReadableSpan sampledSpan = createSampledEndedSpan(SPAN_NAME_2);
    List<SpanData> exported = waitingSpanExporter.waitForExport(1);
    assertThat(exported).containsExactly(sampledSpan.toSpanData());
    */
  }

  @Test(timeout = 10000L)
  public void shutdownFlushes() {
    WaitingSpanExporter waitingSpanExporter = new WaitingSpanExporter(1);
    // Set the export delay to zero, for no timeout, in order to confirm the #flush() below works
    BatchSpansProcessor.Config config =
        BatchSpansProcessor.Config.newBuilder().setScheduleDelayMillis(0).build();

    tracerSdkFactory.addSpanProcessor(BatchSpansProcessor.create(waitingSpanExporter, config));

    ReadableSpan span2 = createSampledEndedSpan(SPAN_NAME_2);

    // Force a shutdown, without this, the #waitForExport() call below would block indefinitely.
    tracerSdkFactory.shutdown();

    List<SpanData> exported = waitingSpanExporter.waitForExport();
    assertThat(exported).containsExactly(span2.toSpanData());
    assertThat(waitingSpanExporter.shutDownCalled.get()).isTrue();
  }

  private static final class BlockingSpanExporter implements SpanExporter {

    final Object monitor = new Object();

    private enum State {
      WAIT_TO_BLOCK,
      BLOCKED,
      UNBLOCKED
    }

    @GuardedBy("monitor")
    State state = State.WAIT_TO_BLOCK;

    @Override
    public ResultCode export(Collection<SpanData> spanDataList) {
      synchronized (monitor) {
        while (state != State.UNBLOCKED) {
          try {
            state = State.BLOCKED;
            // Some threads may wait for Blocked State.
            monitor.notifyAll();
            monitor.wait();
          } catch (InterruptedException e) {
            // Do nothing
          }
        }
      }
      return ResultCode.SUCCESS;
    }

    @Override
    public ResultCode flush() {
      return ResultCode.SUCCESS;
    }

    private void waitUntilIsBlocked() {
      synchronized (monitor) {
        while (state != State.BLOCKED) {
          try {
            monitor.wait();
          } catch (InterruptedException e) {
            // Do nothing
          }
        }
      }
    }

    @Override
    public void shutdown() {
      // Do nothing;
    }

    private void unblock() {
      synchronized (monitor) {
        state = State.UNBLOCKED;
        monitor.notifyAll();
      }
    }
  }

  static class WaitingSpanExporter implements SpanExporter {

    private final List<SpanData> spanDataList = new ArrayList<>();
    private final int numberToWaitFor;
    private CountDownLatch countDownLatch;
    private int timeout = 10;
    private final AtomicBoolean shutDownCalled = new AtomicBoolean(false);

    WaitingSpanExporter(int numberToWaitFor) {
      countDownLatch = new CountDownLatch(numberToWaitFor);
      this.numberToWaitFor = numberToWaitFor;
    }

    WaitingSpanExporter(int numberToWaitFor, int timeout) {
      this(numberToWaitFor);
      this.timeout = timeout;
    }

    /**
     * Waits until we received numberOfSpans spans to export. Returns the list of exported {@link
     * SpanData} objects, otherwise {@code null} if the current thread is interrupted.
     *
     * @return the list of exported {@link SpanData} objects, otherwise {@code null} if the current
     *     thread is interrupted.
     */
    @Nullable
    List<SpanData> waitForExport() {
      try {
        countDownLatch.await(timeout, TimeUnit.SECONDS);
      } catch (InterruptedException e) {
        // Preserve the interruption status as per guidance.
        Thread.currentThread().interrupt();
        return null;
      }
      List<SpanData> result = new ArrayList<>(spanDataList);
      spanDataList.clear();
      return result;
    }

    @Override
    public ResultCode export(Collection<SpanData> spans) {
      this.spanDataList.addAll(spans);
      for (int i = 0; i < spans.size(); i++) {
        countDownLatch.countDown();
      }
      return ResultCode.SUCCESS;
    }

    @Override
    public ResultCode flush() {
      return ResultCode.SUCCESS;
    }

    @Override
    public void shutdown() {
      shutDownCalled.set(true);
    }

    public void reset() {
      this.countDownLatch = new CountDownLatch(numberToWaitFor);
    }
  }
}
